from enum import Enum


class Opt(int, Enum):
    IAC = 255  # "Interpret As Command"
    DONT = 254
    DO = 253
    WONT = 252
    WILL = 251
    NULL = 0

    SE = 240  # Subnegotiation End
    NOP = 241  # No Operation
    DM = 242  # Data Mark
    BRK = 243  # Break
    IP = 244  # Interrupt process
    AO = 245  # Abort output
    AYT = 246  # Are You There
    EC = 247  # Erase Character
    EL = 248  # Erase Line
    GA = 249  # Go Ahead
    SB = 250  # Subnegotiation Begin

    BINARY = 0  # 8-bit data path
    ECHO = 1  # echo
    RCP = 2  # prepare to reconnect
    SGA = 3  # suppress go ahead
    NAMS = 4  # approximate message size
    STATUS = 5  # give status

    LINEMODE = 34



from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Callable, Dict, Generator, Iterable, Iterator, List, Optional, Union

from ._opt import Opt

StateCoroutine = Generator[None, int, None]
BacklogItem = Union["Event", bytes]
EventCallback = Callable[["Event"], Any]
DataCallback = Callable[[bytes], Any]

NUL = 0
LF = 10
CR = 13


@dataclass
class TelnetOption:
    """
    Tracks the status of telnet options
    """

    local_option: Optional[bool] = None
    remote_option: Optional[bool] = None
    reply_pending: bool = False


class Event(ABC):
    @abstractmethod
    def as_bytes(self) -> bytes:  # pragma: nocover
        ...


@dataclass(frozen=True)
class Command(Event):
    cmd: Opt
    opt: Optional[int] = None  # only included for 3-byte commands

    def as_bytes(self) -> bytes:
        data = bytes([Opt.IAC, self.cmd])
        if self.opt is not None:
            data += bytes([self.opt])
        return data


@dataclass(frozen=True)
class SubCommand(Event):
    cmd: Opt
    opts: bytearray

    def as_bytes(self) -> bytes:
        return bytes([self.cmd]) + self.opts


class Backlog(Iterable[BacklogItem]):
    def __init__(self) -> None:
        self.data: List[BacklogItem] = []

    def __iter__(self) -> Iterator[BacklogItem]:
        return iter(self.data)

    def clear(self) -> None:
        self.data.clear()

    def add_message(self, data: bytes) -> None:
        if not self.data or not isinstance(self.data[-1], bytes):
            self.data.append(data)
        else:
            self.data[-1] += data

    def add_event(self, event: Event) -> None:
        self.data.append(event)


class TelnetMachine:
    """
    Sans-IO telnet state machine.
    """

    #: List of IAC commands needing multiple (3) bytes
    _iac_mbs = (Opt.DO, Opt.DONT, Opt.WILL, Opt.WONT)

    def __init__(self) -> None:
        self._state: StateCoroutine = self._state_data()
        next(self._state)  # prime
        self._next_state: Callable[[], StateCoroutine] = self._state_data
        self._backlog = Backlog()
        self._leftover = bytearray()
        self._event_callbacks: List[EventCallback] = []
        self._receive_callbacks: List[DataCallback] = []
        self._send_callbacks: List[DataCallback] = []

    def register_event_cb(self, callback: EventCallback) -> None:
        """
        Register a callback for every event created.
        """
        self._event_callbacks.append(callback)

    def register_receive_cb(self, callback: DataCallback) -> None:
        """
        Register a callback for message data that was received.
        """
        self._receive_callbacks.append(callback)

    def register_send_cb(self, callback: DataCallback) -> None:
        """
        Register a callback for raw data that needs to be sent.
        """
        self._send_callbacks.append(callback)

    def receive_data(self, data: bytes) -> None:
        """
        Receive data from a stream into the state machine.
        """
        data = self._leftover + data
        self._leftover.clear()
        for idx, byte in enumerate(data):
            try:
                self._state.send(byte)
            except StopIteration:
                self._state = self._next_state()
                next(self._state)  # prime
            except:
                self._leftover.extend(data[idx:])
                raise

        for event_or_data in self._backlog:
            if isinstance(event_or_data, bytes):
                for recv_cb in self._receive_callbacks:
                    recv_cb(event_or_data)
            else:
                for event_cb in self._event_callbacks:
                    event_cb(event_or_data)
        self._backlog.clear()

    def send(self, data: bytes) -> None:
        """
        Send raw data to the stream.
        """
        for cb in self._send_callbacks:
            cb(data)

    def send_message(self, data: bytes) -> None:
        """
        Send a message (text) to the stream.
        """
        data = data.replace(bytes([Opt.IAC]), bytes([Opt.IAC, Opt.IAC]))
        self.send(data)

    def send_command(self, cmd: Opt, opt: Optional[int] = None) -> None:
        self.send(Command(cmd, opt).as_bytes())

    def _state_data(self) -> StateCoroutine:
        while True:
            char = yield

            if char == Opt.IAC:
                self._next_state = self._state_iac
                return
            elif char == CR:
                char_after_cr = yield

                if char_after_cr == LF:
                    self._backlog.add_message(b"\n")
                elif char_after_cr == NUL:
                    self._backlog.add_message(b"\r")
                elif char_after_cr == Opt.IAC:
                    # IAC isn't allowed after \r according to the
                    # RFC, but handling this way is less surprising than
                    # delivering the IAC as normal data.
                    self._backlog.add_message(b"\r")
                    self._next_state = self._state_iac
                    return
                else:
                    self._backlog.add_message(b"\r" + bytes([char_after_cr]))
            else:
                self._backlog.add_message(bytes([char]))

    def _state_iac(self) -> StateCoroutine:
        self._next_state = self._state_data

        cmd = yield
        if cmd in self._iac_mbs:
            # These commands are 3-byte
            opt = yield
            self._backlog.add_event(Command(Opt(cmd), opt))
        elif cmd == Opt.IAC:
            # escaped IAC
            self._backlog.add_message(bytes([Opt.IAC]))
        elif cmd == Opt.SB:
            self._next_state = self._state_sb
        else:
            self._backlog.add_event(Command(Opt(cmd)))

    def _state_sb(self) -> StateCoroutine:
        buf = bytearray()
        self._next_state = self._state_data

        while True:
            data = yield
            if data == Opt.IAC:
                opt = yield
                if opt == Opt.SE:
                    if not buf:
                        raise ValueError("SE: buffer empty")
                    if buf[0] == Opt.NULL:
                        raise ValueError("SE: buffer is NUL")
                    if len(buf) == 1:
                        raise ValueError("SE: buffer too short")

                    cmd = Opt(buf[0])
                    self._backlog.add_event(SubCommand(cmd, buf[1:]))
                    break
                elif opt == Opt.IAC:
                    # escaped IAC
                    buf.append(Opt.IAC)
            else:
                buf.append(data)


class TelnetClient:
    def __init__(self, machine: TelnetMachine) -> None:
        self._machine = machine
        self._machine.register_event_cb(self.on_event)
        self._options: Dict[int, TelnetOption] = defaultdict(TelnetOption)

    def on_event(self, event: Event) -> None:
        if isinstance(event, Command):
            if event.opt is None:
                raise ValueError()
            if event.cmd == Opt.DO:
                return self.handle_do(event.opt)
            elif event.cmd == Opt.DONT:
                return self.handle_dont(event.opt)
            elif event.cmd == Opt.WILL:
                return self.handle_will(event.opt)
            elif event.cmd == Opt.WONT:
                return self.handle_wont(event.opt)

    def check_local_option(self, option: Opt) -> bool:
        return self._options[option].local_option or False

    def check_remote_option(self, option: Opt) -> bool:
        return self._options[option].remote_option or False

    def handle_do(self, option: int) -> None:
        self._options[option].reply_pending = False

        if option == Opt.ECHO:
            # DE requests us to echo their input
            if not self._options[Opt.ECHO].local_option:
                self._options[Opt.ECHO].local_option = True
                self._machine.send_command(Opt.WILL, Opt.ECHO)

        elif option == Opt.BINARY:
            # DE requests us to receive BINARY
            if not self._options[Opt.BINARY].local_option:
                self._options[Opt.BINARY].local_option = True
                self._machine.send_command(Opt.WILL, Opt.BINARY)

        elif option == Opt.SGA:
            # DE wants us to suppress go-ahead
            if not self._options[Opt.SGA].local_option:
                self._options[Opt.SGA].local_option = True
                self._machine.send_command(Opt.WILL, Opt.SGA)
                self._machine.send_command(Opt.DO, Opt.SGA)

    def handle_dont(self, option: int) -> None:
        pass

    def handle_will(self, option: int) -> None:
        pass

    def handle_wont(self, option: int) -> None:
        pass


class TelnetServer:
    def __init__(self, machine: TelnetMachine) -> None:
        self._machine = machine
        self._options: Dict[int, TelnetOption] = defaultdict(TelnetOption)



from ._machine import Command, Event, SubCommand, TelnetClient, TelnetMachine, TelnetServer
from ._opt import Opt

__all__ = [
    "Command",
    "Event",
    "SubCommand",
    "Opt",
    "TelnetClient",
    "TelnetMachine",
    "TelnetServer",
]

try:
    import anyio
except ImportError:  # pragma: nocover
    pass
else:
    from ._anyio import AnyioTelnetClient, AnyioTelnetServer

    __all__ += [
        "AnyioTelnetClient",
        "AnyioTelnetServer",
    ]



import math
import sys
from types import TracebackType
from typing import Any, AsyncIterator, Callable, Optional, Type, TypeVar

import anyio
from anyio.abc import AnyByteStream, ByteStream, TaskGroup
from anyio.streams.buffered import BufferedByteReceiveStream

from ._machine import TelnetClient, TelnetMachine, TelnetServer
from ._opt import Opt

if sys.version_info >= (3, 7):
    from contextlib import asynccontextmanager
else:
    from async_generator import asynccontextmanager

T = TypeVar("T", bound="_AnyioTelnet")


class _AnyioTelnet(ByteStream):
    _task_group: TaskGroup

    def __init__(self, stream: AnyByteStream, on_receive_error: Optional[Callable[[Exception], Any]] = None) -> None:
        self._stream = stream
        self._on_receive_error = on_receive_error

        self._msg_stream_producer, msg_stream_consumer = anyio.create_memory_object_stream(math.inf, item_type=bytes)
        self._msg_stream_buff = BufferedByteReceiveStream(msg_stream_consumer)
        send_producer, self._send_consumer = anyio.create_memory_object_stream(math.inf, item_type=bytes)

        self._machine = TelnetMachine()
        self._machine.register_receive_cb(self._msg_stream_producer.send_nowait)
        self._machine.register_send_cb(send_producer.send_nowait)

    @asynccontextmanager
    async def _make_ctx(self: T) -> AsyncIterator[T]:
        async with anyio.create_task_group() as self._task_group:
            self._task_group.start_soon(self._receive_worker)
            self._task_group.start_soon(self._send_worker)
            try:
                yield self
            finally:
                self._task_group.cancel_scope.cancel()
                await self.aclose()

    async def __aenter__(self: T) -> T:
        self._ctx = self._make_ctx()
        return await self._ctx.__aenter__()

    async def __aexit__(
        self, exc_type: Optional[Type[BaseException]], exc: Optional[BaseException], tb: Optional[TracebackType]
    ) -> None:
        await self._ctx.__aexit__(exc_type, exc, tb)

    async def _receive_worker(self) -> None:
        try:
            async for data in self._stream:
                try:
                    self._machine.receive_data(data)
                except anyio.get_cancelled_exc_class():
                    raise
                except Exception as exc:
                    if self._on_receive_error is not None:
                        self._on_receive_error(exc)
                    else:
                        raise
        finally:
            await self.aclose()

    async def _send_worker(self) -> None:
        async for data in self._send_consumer:
            await self._stream.send(data)

    async def aclose(self) -> None:
        await self._msg_stream_producer.aclose()

    async def send_eof(self) -> None:
        await self._stream.send_eof()

    async def receive(self, max_bytes: int = 4096) -> bytes:
        return await self._msg_stream_buff.receive(max_bytes=max_bytes)

    async def send(self, data: bytes) -> None:
        self._machine.send_message(data)

    async def send_command(self, cmd: Opt, opt: Optional[Opt] = None) -> None:
        self._machine.send_command(cmd, opt)


class AnyioTelnetServer(_AnyioTelnet):
    def __init__(self, stream: AnyByteStream, *args: Any, **kwargs: Any) -> None:
        super().__init__(stream, *args, **kwargs)
        self._server = TelnetServer(self._machine)


class AnyioTelnetClient(_AnyioTelnet):
    def __init__(self, stream: AnyByteStream, *args: Any, **kwargs: Any) -> None:
        super().__init__(stream, *args, **kwargs)
        self._client = TelnetClient(self._machine)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['telnetio']

package_data = \
{'': ['*']}

extras_require = \
{':python_version == "3.6"': ['dataclasses>=0.8,<0.9'],
 'anyio': ['anyio>=3.2.0,<4.0.0'],
 'anyio:python_version == "3.6"': ['async_generator>=1.10,<2.0']}

setup_kwargs = {
    'name': 'telnetio',
    'version': '0.3.0',
    'description': 'Sans-IO telnet parser',
    'long_description': 'telnetio\n========\n\nA Sans-IO implementation of a telnet parser.\n\nIncludes an `anyio` server implementation.  To install use the `anyio` extra, such as `pip install telnetio[anyio]`.\n\n```python\nimport anyio\nfrom anyio.abc import AnyByteStream\n\nfrom telnetio import AnyioTelnetServer\n\nasync def handler(stream: AnyByteStream) -> None:\n    async with stream, AnyioTelnetServer(stream) as telnet:\n        async for data in telnet:\n            await telnet.send(data)\n\nasync def main() -> None:\n    listener = await anyio.create_tcp_listener(local_port=1234)\n    await listener.serve(handler)\n\nanyio.run(main)\n```\n\nSee the `examples` directory for more examples.\n',
    'author': 'Jordan Speicher',
    'author_email': 'jordan@jspeicher.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/uSpike/telnetio',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)
